from sklearn.cluster import KMeans
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axisartist.axislines import SubplotZero
from sklearn import manifold
from matplotlib.colors import LinearSegmentedColormap
import os
import matplotlib.image as imgplt





class FindNiche():

    def __init__(self, output_path):

        """
        FindNiche.

        Parameters
        ----------
        output_path
            Path of the output folder.

        """

        self.name = "FindNiche"
        self.output_path = output_path

        if not os.path.exists(self.output_path + "/Downstream/"):
            os.makedirs(self.output_path + "/Downstream/")



    def Kmeans(self, predicted_raw, predicted_imputed, n_kmeans=8):

        """
        Visualization of loss curve and accuracy curve.

        Parameters
        ----------
        predicted_raw
            Path of predicted raw spots with detailed probabilities.
        predicted_imputed
            Path of predicted imputed spots with detailed probabilities.
        n_kmeans
            Numbers of clusters generated by Kmeans.
        """

        predict = []

        with open(predicted_raw) as f:
            for line in f.readlines():
                temp = line.split()
                predict.append(temp)
        print(len(predict))

        with open(predicted_imputed) as f:
            for line in f.readlines():
                temp = line.split()
                predict.append(temp)
        print(len(predict))

        data = []
        for each in predict:
            data.append(each[1:-2])

        data = np.array(data, dtype='float')

        kmeans = KMeans(n_clusters=n_kmeans).fit(data)
        n_labels = kmeans.labels_

        for i in range(len(predict)):
            predict[i].append(n_labels[i])

        np.savetxt(self.output_path + "/Downstream/kmeans_label.txt", predict, fmt='%s', delimiter='\t')



    def TSNE_Plot(self, path, kmeans_file, Label, colors):


        """
        Visualization of TSNE plot.

        Parameters
        ----------
        path
            Path of predicted files with detailed probabilities.
        kmeans_file
            Path of files with kmeans labels.
        Label
            Dictionary of predicted labels.
        colormap
            Dictionary of colors for labels.
        """

        barcode, predict_data, cell_label, kmeans_label = [], [], [], []

        with open(path + "/Imputedpredict_detail.txt") as f:
            for line in f.readlines():
                temp = line.split()
                barcode.append(temp[0])
                predict_data.append(temp[1:-2])

        with open(path + "/Imputed_predict.txt") as f:
            for line in f.readlines():
                temp = line.split()
                cell_label.append(temp[1])

        with open(path + "/Rawpredict_detail.txt") as f:
            for line in f.readlines():
                temp = line.split()
                barcode.append(temp[0])
                predict_data.append(temp[1:-2])

        with open(path + "/Raw_predict.txt") as f:
            for line in f.readlines():
                temp = line.split()
                cell_label.append(temp[1])

        with open(kmeans_file) as f:
            for line in f.readlines():
                temp = line.split()
                kmeans_label.append(temp[-1])

        data = np.array(predict_data, dtype='float')
        clabel = np.array(cell_label, dtype='int')
        klabel = np.array(kmeans_label, dtype='int')

        tsne = manifold.TSNE(n_components=2, random_state=24)
        embedding = tsne.fit_transform(data)
        print(embedding.shape)

        true_label = np.unique(clabel)
        print(true_label)

        colorslist = []
        for each in Label:
            colorslist.append(colors[Label[each]])
        newcmp = LinearSegmentedColormap.from_list('mylist', colorslist)

        plt.scatter(embedding[:, 0], embedding[:, 1], c=clabel, cmap=newcmp, s=5, vmin=0, vmax=len(Label) - 1)
        plt.gca().set_aspect('equal', 'datalim')
        plt.colorbar(boundaries=np.arange(len(Label) + 1) - 0.5).set_ticks(np.arange(len(Label)))
        plt.title('tSNE ' + str(true_label))
        plt.savefig(self.output_path + "/Downstream/tsne_celltype.pdf")
        plt.clf()

        plt.scatter(embedding[:, 0], embedding[:, 1], c=klabel, cmap="Accent", s=5)
        plt.gca().set_aspect('equal', 'datalim')
        plt.colorbar()
        plt.title('tSNE')
        plt.savefig(self.output_path + "/Downstream/tsne_kmeans.pdf")
        plt.clf()



    def SpatialNiche(self, kmeans_file, image_path, colormap, pointsize=5, back_image=False):

        """
        Spatial plot of cell-type niches.

        Parameters
        ----------
        kmeans_file
            Path of files with kmeans labels.
        image_path
            Path of the original H&E staining image.
        colormap
            Dictionary of colors for labels.
        output_path
            Path of output files.
        pointsize
            Size of scatter points.
        back_image
            Shown of the original H&E staining image.
        """

        Image = imgplt.imread(image_path)
        h, w, _ = Image.shape

        predict = []
        with open(kmeans_file) as f:
            for line in f.readlines():
                temp = line.split()
                predict.append(temp)

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)
        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)

        if back_image == True:
            plt.imshow(Image)

        for each in predict:
            plt.scatter(float(each[-3]), float(each[-2]), c=colormap[int(each[-1])], s=pointsize, marker='o')
        plt.savefig(self.output_path + "/Downstream/kmeans_spatial.pdf")










import tensorflow as tf

gpus = tf.config.list_physical_devices('GPU')
for gpu in gpus:
    tf.config.experimental.set_memory_growth(device=gpu, enable=True)
    
from tensorflow.keras.applications.vgg16 import VGG16
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D, MaxPool2D, Activation, Dropout, Flatten, Dense
from tensorflow.keras.optimizers import SGD
from tensorflow.keras.preprocessing.image import ImageDataGenerator, img_to_array, load_img

from tensorflow.keras.models import load_model
import numpy as np
import os

import itertools
import glob





class BuildModel():

    def __init__(self):

        self.name = "BuildModel"

        self.train_datagen = ImageDataGenerator(
            rotation_range=40,
            width_shift_range=0.2,
            height_shift_range=0.2,
            rescale=1 / 255,
            shear_range=20,
            zoom_range=0.2,
            horizontal_flip=True,
            fill_mode='nearest',
        )

        self.test_datagen = ImageDataGenerator(
            rescale=1 / 255,
        )

        self.batch_size = 32



    def base_model(self, traindata_path, testdata_path, output_path, n_class, lr=1e-3, epochs=50, model_name="base_model.h5"):

        """
        Build base model.

        Parameters
        ----------
        traindata_path
            Path of training set.
        testdata_path
            Path of testing set.
        output_path
            Path of output files.
        n_class
            Number of cell classes.
        lr
            Learning rate for SGDOptimizer.
        epochs
            Number of total epochs in training.
        model_name
            Name of trained model
        """

        vgg16_model = VGG16(weights='imagenet', include_top=False, input_shape=(224, 224, 3))

        top_model = Sequential()
        top_model.add(Flatten(input_shape=vgg16_model.output_shape[1:]))
        top_model.add(Dense(256, activation='relu'))
        top_model.add(Dropout(0.5))
        top_model.add(Dense(n_class, activation='softmax'))
        model = Sequential()
        model.add(vgg16_model)
        model.add(top_model)

        train_generator = self.train_datagen.flow_from_directory(
            traindata_path,
            target_size=(224, 224),
            batch_size=self.batch_size,
        )

        test_generator = self.test_datagen.flow_from_directory(
            testdata_path,
            target_size=(224, 224),
            batch_size=self.batch_size,
        )

        print(train_generator.class_indices)
        print(test_generator.class_indices)


        model.compile(optimizer=SGD(lr=lr, momentum=0.9), loss='categorical_crossentropy', metrics=['accuracy'])
        model.fit(train_generator, epochs=epochs, validation_data=test_generator)

        loss, val_loss, accuracy, val_accuracy = model.history.history['loss'], model.history.history['val_loss'], model.history.history['accuracy'], model.history.history['val_accuracy']

        os.makedirs(output_path + "/Models/")

        log = [loss, val_loss, accuracy, val_accuracy]
        np.savetxt(output_path + "/Models/" + "Log_BaseModel.txt", log, fmt='%s', delimiter='\t')

        model.save(output_path + "/Models/" + model_name)



    def finetuned_model(self, basemodel, traindata_path, testdata_path, output_path, lr=1e-4, epochs=50, model_name="finetuned_model.h5"):

        """
        Build finetuned model.

        Parameters
        ----------
        basemodel
            Path of base model
        traindata_path
            Path of training set.
        testdata_path
            Path of testing set.
        output_path
            Path of output files.
        lr
            Learning rate for SGDOptimizer.
        epochs
            Number of total epochs in training.
        model_name
            Name of trained model
        """

        model = load_model(basemodel)

        train_generator = self.train_datagen.flow_from_directory(
            traindata_path,
            target_size=(224, 224),
            batch_size=self.batch_size,
        )

        test_generator = self.test_datagen.flow_from_directory(
            testdata_path,
            target_size=(224, 224),
            batch_size=self.batch_size,
        )

        print(train_generator.class_indices)
        print(test_generator.class_indices)

        model.compile(optimizer=SGD(lr=lr, momentum=0.9), loss='categorical_crossentropy', metrics=['accuracy'])
        model.fit(train_generator, epochs=epochs, validation_data=test_generator)

        loss, val_loss, accuracy, val_accuracy = model.history.history['loss'], model.history.history['val_loss'], model.history.history['accuracy'], model.history.history['val_accuracy']

        if not os.path.exists(output_path + "/Models/"):
            os.makedirs(output_path + "/Models/")

        log = [loss, val_loss, accuracy, val_accuracy]
        np.savetxt(output_path + "/Models/" + "Log_FinetunedModel.txt", log, fmt='%s', delimiter='\t')

        model.save(output_path + "/Models/" + model_name)



    def alltrain_model(self, traindata_path, output_path, n_class, lr=1e-3, epochs=50, model_name="alltrain_model.h5"):

        """
        Build base model.

        Parameters
        ----------
        traindata_path
            Path of training set.
        output_path
            Path of output files.
        n_class
            Number of cell classes.
        lr
            Learning rate for SGDOptimizer.
        epochs
            Number of total epochs in training.
        model_name
            Name of trained model
        """

        vgg16_model = VGG16(weights='imagenet', include_top=False, input_shape=(224, 224, 3))

        top_model = Sequential()
        top_model.add(Flatten(input_shape=vgg16_model.output_shape[1:]))
        top_model.add(Dense(256, activation='relu'))
        top_model.add(Dropout(0.5))
        top_model.add(Dense(n_class, activation='softmax'))
        model = Sequential()
        model.add(vgg16_model)
        model.add(top_model)

        train_generator = self.train_datagen.flow_from_directory(
            traindata_path,
            target_size=(224, 224),
            batch_size=self.batch_size,
        )


        print(train_generator.class_indices)


        model.compile(optimizer=SGD(lr=lr, momentum=0.9), loss='categorical_crossentropy', metrics=['accuracy'])
        model.fit(train_generator, epochs=epochs)

        os.makedirs(output_path + "/Models/")
        model.save(output_path + "/Models/" + model_name)





class BuildPrediction():

    def __init__(self, model, result_dir):

        """
        Build prediction.

        Parameters
        ----------
        model
            Path of the trained model used for prediction.
        result_dir
            Path of the output folder.
        """

        self.name = "Prediction"
        self.model = load_model(model)
        self.result_dir = result_dir

        if not os.path.exists(result_dir + "/Predict/"):
            os.makedirs(result_dir + "/Predict/")



    def predict(self, image):

        """
        Build prediction.

        Parameters
        ----------
        image
            Image for prediction.
        """

        image = load_img(image)
        image = image.resize((224, 224))
        image = img_to_array(image)
        image = image / 255
        image = np.expand_dims(image, 0)
        predict = self.model.predict(image)
        classes = np.argmax(predict, axis=1)

        return predict, classes



    def prediction(self, input_spot, input_image, input_position, predicting_type):

        """
        Build prediction.

        Parameters
        ----------
        input_spot
            Spots in the dataset, which used to determine labels.
        input_image
            Path of images for prediction.
        input_position
            Position file of images.
        predicting_type
            Type of predicting spot images.
        """

        celltype = list(set([x[-1] for x in input_spot]))
        celltype.sort()
        label = {}
        for i in range(len(celltype)):
            label[i] = celltype[i]
        print(label)

        imgsLib = []
        imgsLib.extend(glob.glob(os.path.join(input_image + "/", "*.png")))

        pre_type = []
        pre_type_detail = []

        for each in imgsLib:
            pre_detail, pre = self.predict(each)
            imgname = each.split("/")[-1].split(".")[0]
            pre_type.append([imgname, int(pre), label[int(pre)]])
            pre_temp = list(itertools.chain.from_iterable(pre_detail.tolist()))
            temp = [imgname]
            temp.extend(pre_temp)
            pre_type_detail.append(temp)

        coor = {}
        with open(input_position) as f:
            for line in f.readlines():
                temp = line.split()
                coor[temp[2]] = [temp[0], temp[1]]

        filltype = []
        for each in pre_type:
            if each[0] in coor.keys():
                temp = each
                temp.extend(coor[each[0]])
                filltype.append(temp)
        np.savetxt(self.result_dir + "/Predict/" + predicting_type + "_predict.txt", filltype, fmt='%s', delimiter='\t')

        filltype_detail = []
        for each in pre_type_detail:
            if each[0] in coor.keys():
                temp = each
                temp.extend(coor[each[0]])
                filltype_detail.append(temp)
        np.savetxt(self.result_dir + "/Predict/" + predicting_type + "predict_detail.txt", filltype_detail, fmt='%s', delimiter='\t')




#!/usr/bin/env python

"""
# Author: Ying Wu
# File Name: __init__.py
# Description:
"""

__author__ = "Ying Wu"
__email__ = "wuy@big.ac.cn"
__name__ = "STASCAN"

#import run_STASCAN
#import preparation
#import model
#import stat
#import downstream

#from STASCAN import run_STASCAN
#from STASCAN import preparation
#from STASCAN import model
#from STASCAN import StatPlot
#from STASCAN import downstream

import STASCAN.run_STASCAN as run_STASCAN
import STASCAN.preparation as preparation
import STASCAN.model as model
import STASCAN.StatPlot as StatPlot
import STASCAN.downstream as downstream


from pylab import *
import numpy as np
import cv2
import os
import random
import glob
import shutil
from PIL import Image





class Generator():

    def __init__(self, record_dir, position, image, crop_size):

        """
        Preparation of spot images.

        Parameters
        ----------
        record_dir
            Path of the output folder.
        position
            The position of raw spots. Standard output file for Space Ranger.
        image
            The original H&E staining images, used for Space Ranger previously.
        crop_size
            Pixel radius of spot images.
        """

        self.name = "Generator"

        self.record_dir = record_dir

        self.position = []
        with open(position) as f:
            for line in f.readlines():
                temp = line.split(',')
                if int(temp[1]) == 1:
                    self.position.append([int(temp[-1].split('\n')[0]), int(temp[-2]), temp[0]])
        self.position.sort(key=lambda x: (x[1], x[0]))
        print("The number of raw spots : ", len(self.position))
        
        self.Data = {}
        for each in self.position:
            self.Data[each[2]] = [each[0], each[1]]

        self.rawimage = image
        self.crop_size = crop_size



    def SpotImage_generator(self, type):

        """
        Generator of spot images.

        Parameters
        ----------
        type: {'RawSpot', 'PriorSpot', 'ImputedSpot'}
            Type of spots.
        """

        img = cv2.imread(self.rawimage)
        crop_size = self.crop_size

        if type == 'RawSpot':
            position = self.position
            for each in position:
                k = each[2].replace("\"", "")
                cropped = img[int(each[1]) - crop_size:int(each[1]) + crop_size,
                          int(each[0]) - crop_size:int(each[0]) + crop_size]
                cv2.imwrite(self.record_dir + "/RawSpot/" + k + ".png", cropped)
        elif type == 'PriorSpot':
            position = self.prior_spot
            for each in position:
                k = each[0].replace("\"", "")
                cropped = img[int(each[2]) - crop_size:int(each[2]) + crop_size,
                          int(each[1]) - crop_size:int(each[1]) + crop_size]
                celltype = each[-1]
                cv2.imwrite(self.record_dir + "/PriorSpot/" + celltype + "-" + k + ".png", cropped)
        elif type == 'ImputedSpot':
            position = self.imputed_spot
            for each in position:
                k = each[2].replace("\"", "")
                each[0], each[1] = float(each[0]), float(each[1])
                cropped = img[int(each[1]) - crop_size:int(each[1]) + crop_size,
                          int(each[0]) - crop_size:int(each[0]) + crop_size]
                cv2.imwrite(self.record_dir + "/ImputedSpot/" + k + ".png", cropped)
        else:
            print("ERROR :The type of SpotImage doesn't exist!")



    def SubSpotImage_generator(self, type):

        """
        Generator of subdivided spot images.

        Parameters
        ----------
        type: {'raw_divided', 'prior_divided'}
            Type of spots.
        """

        imgsLib = []
        subloc = []
        position = self.Data

        if type == 'raw_divided':
            img_path = self.record_dir + "/RawSpot/"
            save_path = self.record_dir + "/SubSpot/raw_divided/"
            #position = self.Data
        if type == 'prior_divided':
            img_path = self.record_dir + "/PriorSpot/"
            save_path = self.record_dir + "/SubSpot/prior_divided/"
            #position = {}
            #if self.prior_spot == True:
            #    for each in self.prior_spot:
            #        position[each[0]] = [int(each[1]), int(each[2])]
            #else:
            #with open(self.record_dir + "/PriorSpot/prior_spot.txt") as f:
            #    for line in f.readlines():
            #        temp = line.split()
            #        position[temp[0]] = [float(temp[1]), float(temp[2])]          
        
        

        imgsLib.extend(glob.glob(os.path.join(img_path, "*.png")))

        for each in imgsLib:
            # img
            image = Image.open(each)
            img = image.convert('RGB')
            w = img.size[0]
            h = img.size[1]
            rawname = each.split("/")[-1].split(".")[0]
            img_1 = img.crop([0, 0, w / 2, h / 2])
            img_1.save(save_path + rawname + "_1.png")
            img_2 = img.crop([w / 2, 0, w, h / 2])
            img_2.save(save_path + rawname + "_2.png")
            img_3 = img.crop([0, h / 2, w / 2, h])
            img_3.save(save_path + rawname + "_3.png")
            img_4 = img.crop([w / 2, h / 2, w, h])
            img_4.save(save_path + rawname + "_4.png")
            # loc
            rawname = each.split("/")[-1].split(".")[0].split("-")[-2] + "-" + each.split("/")[-1].split(".")[0].split("-")[-1]
            rawxy = position[rawname]
            subloc_1 = [rawxy[0] - self.crop_size / 2, rawxy[1] - self.crop_size / 2, rawname + '_1']
            subloc.append(subloc_1)
            subloc_2 = [rawxy[0] + self.crop_size / 2, rawxy[1] - self.crop_size / 2, rawname + '_2']
            subloc.append(subloc_2)
            subloc_3 = [rawxy[0] - self.crop_size / 2, rawxy[1] + self.crop_size / 2, rawname + '_3']
            subloc.append(subloc_3)
            subloc_4 = [rawxy[0] + self.crop_size / 2, rawxy[1] + self.crop_size / 2, rawname + '_4']
            subloc.append(subloc_4)

        newsubloc = np.array(subloc)
        np.savetxt(save_path + "subloc.txt", newsubloc, fmt='%s', delimiter='\t')



    def PriorSpot_generator(self, prelabel_path, strategies='joint', threshold_proportion = 0.6, threshold_ration = 1.5):

        """
        Generator of prior spot.

        Parameters
        ----------
        prelabel_path
            Path of prelabel files.
        strategies: {'joint', 'single_proportion', 'single_ration'}
            Type of strategies to label the cell types for each spot.
        threshold_proportion
            The threshold of prior spot selection. Spots which the proportion of dominant cell types reached the threshold were selected as prior spots.
        threshold_ration
            The threshold of prior spot selection. Spots which the proportion of dominant cell types exceeded the given multiple of the proportion of secondary cell types were selected as prior spots.
        """

        self.prelabel_strategies = strategies
        self.threshold_proportion = threshold_proportion
        self.threshold_ration = threshold_ration
        print("The strategy of pre-labelling is: " + str(self.prelabel_strategies))

        if not os.path.exists(self.record_dir+"/PriorSpot/"):
            os.makedirs(self.record_dir+"/PriorSpot/")

        #self.Data = {}
        #for each in self.position:
        #    self.Data[each[2]] = [each[0], each[1]]

        spot_C = {}
        with open(prelabel_path + "Cell2location_prelabelling.txt") as f:
            for line in f.readlines():
                temp = line.split()
                spot_C[temp[0]] = temp[1:]
        with open(prelabel_path + "Cell2location_prelabelling.txt") as f:
            Label_C = f.readlines()[0].split()

        prior_spot = []

        if self.prelabel_strategies == 'joint':
            spot_S = {}
            with open(prelabel_path + "Seurat_prelabelling.txt") as f:
                for line in f.readlines():
                    temp = line.split()
                    spot_S[temp[0]] = temp[1:]
            with open(prelabel_path + "Seurat_prelabelling.txt") as f:
                Label_S = f.readlines()[0].split()

            for each in self.Data.keys():
                if each in spot_C.keys() and each in spot_S.keys():
                    value_C = [float(k) for k in spot_C[each]]
                    maxtype_C = Label_C[value_C.index(max(value_C))]
                    value_S = [float(k) for k in spot_S[each]]
                    maxtype_S = Label_S[value_S.index(max(value_S))]
                    if maxtype_C == maxtype_S:
                        temp = [each]
                        temp.extend(self.Data[each])
                        temp.extend([maxtype_C, maxtype_S])
                        prior_spot.append(temp)

        if self.prelabel_strategies == 'single_proportion':
            print("The threshold of cell proportion is: " + str(threshold_proportion))
            for each in self.Data.keys():
                if each in spot_C.keys():
                    value_C = [float(k) for k in spot_C[each]]
                    maxtype_C = Label_C[value_C.index(max(value_C))]
                    if max(value_C) / sum(value_C) > threshold_proportion:
                        temp = [each]
                        temp.extend(self.Data[each])
                        temp.extend([maxtype_C, maxtype_C])
                        prior_spot.append(temp)

        if self.prelabel_strategies == 'single_ration':
            print("The threshold of cell ration is: " + str(threshold_ration))
            for each in self.Data.keys():
                if each in spot_C.keys():
                    value_C = [float(k) for k in spot_C[each]]
                    ctype = Label_C[value_C.index(max(value_C))]
                    value_C.sort()
                    if value_C[-1] > value_C[-2] * threshold_ration:
                        temp = [each]
                        temp.extend(self.Data[each])
                        temp.extend([ctype, ctype])
                        prior_spot.append(temp)

        self.prior_spot = prior_spot
        print("The number of prior spots : ", len(prior_spot))
        np.savetxt(self.record_dir+"/PriorSpot/prior_spot.txt", prior_spot, fmt='%s', delimiter='\t')
        self.SpotImage_generator("PriorSpot")

        return prior_spot



    def RawSpot_generator(self):

        if not os.path.exists(self.record_dir + "/RawSpot/"):
            os.makedirs(self.record_dir + "/RawSpot/")

        self.SpotImage_generator("RawSpot")



    def SubSpot_generator(self):

        if not os.path.exists(self.record_dir + "/SubSpot/"):
            os.makedirs(self.record_dir + "/SubSpot/")
        if not os.path.exists(self.record_dir + "/SubSpot/raw_divided/"):
            os.makedirs(self.record_dir + "/SubSpot/raw_divided/")
        if not os.path.exists(self.record_dir + "/SubSpot/prior_divided/"):
            os.makedirs(self.record_dir + "/SubSpot/prior_divided")

        self.SubSpotImage_generator("raw_divided")
        self.SubSpotImage_generator("prior_divided")



    def ImputedSpot_generator(self):

        if not os.path.exists(self.record_dir + "/ImputedSpot/"):
            os.makedirs(self.record_dir + "/ImputedSpot/")

        all_s = []
        sess = []
        s = int(self.position[0][1])

        for i in range(len(self.position)):
            if self.position[i][1] == s:
                sess.append(self.position[i])
            else:
                if int(self.position[i][1]) < s + self.crop_size:
                    sess.append(self.position[i])
                else:
                    maxv = int(self.position[i - 1][1])
                    averv = (s + maxv) / 2
                    for j in range(len(sess)):
                        sess[j][1] = averv
                    all_s.append(sess)
                    sess = []
                    s = self.position[i][1]
                    sess.append(self.position[i])

        maxv = int(self.position[-1][1])
        averv = (s + maxv) / 2
        for j in range(len(sess)):
            sess[j][1] = averv
        all_s.append(sess)

        for each in all_s:
            each.sort(key=lambda x: (x[0]))

        alls = []
        for i in range(len(all_s)):
            for j in range(len(all_s[i])):
                alls.append(all_s[i][j])

        np.savetxt(self.record_dir + "/ImputedSpot/adjust_raw_spot.txt", alls, fmt='%s', delimiter='\t')

        fill = []

        for i in range(len(all_s)):
            for j in range(len(all_s[i]) - 1):
                a = int(all_s[i][j + 1][0]) - int(all_s[i][j][0])
                nx = int(all_s[i][j][0]) + a / 2
                ny = int(all_s[i][j][1]) + int(a / ((3 ** 0.5) * 2))
                nl = 'new-down' + all_s[i][j][2]
                fill.append(nx)
                fill.append(ny)
                fill.append(nl)
                ny = int(all_s[i][j][1]) - int(a / ((3 ** 0.5) * 2))
                nl = 'new-up' + all_s[i][j][2]
                fill.append(nx)
                fill.append(ny)
                fill.append(nl)

        newfill = np.array(fill)
        newfill = newfill.reshape(int(len(newfill) / 3), 3)

        np.savetxt(self.record_dir + "/ImputedSpot/imputed_spot.txt", newfill, fmt='%s', delimiter='\t')
        print("The number of imputed spots : ", len(newfill))

        self.imputed_spot = newfill.tolist()
        self.SpotImage_generator("ImputedSpot")





class Shuffler():

    def __init__(self, record_dir):

        """
        Shuffler of spot images.

        Parameters
        ----------
        record_dir
            Path of the input folder.
        """

        self.name = "Shuffler"
        self.record_dir = record_dir



    def Dataset_divider(self, input_spot, shuffle_scale=0.2):

        """
        Divider of spot image dataset.

        Parameters
        ----------
        input_spot
            Spots in the dataset, which used to determine cell classes.
        shuffle_scale
            Shuffle scale between training and testing sets.
        """

        Dir = list(set([x[-1] for x in input_spot]))

        os.makedirs(self.record_dir + "/train/")
        os.makedirs(self.record_dir + "/test/")
        for each in Dir:
            os.makedirs(self.record_dir + "/train/" + each)
            os.makedirs(self.record_dir + "/test/" + each)

        imgsLib = []
        for n in Dir:
            imgsLib.extend(glob.glob(os.path.join(self.record_dir, n + "*.png")))
            for each in imgsLib:
                dirpath = self.record_dir + "/train/" + n + "/"
                shutil.move(each, dirpath)
            imgsLib = []

        imgsLib = []
        for n in Dir:
            imgsLib.extend(glob.glob(os.path.join(self.record_dir + "/train/" + n + "/", "*.png")))
            np.random.shuffle(imgsLib)
            num = int(len(imgsLib) * shuffle_scale)
            for i in range(len(imgsLib[:num])):
                scrpath = imgsLib[i]
                dispath = self.record_dir + "/test/" + n + "/"
                shutil.move(scrpath, dispath)
            imgsLib = []



    def KCross_divider(self, K_group, output_path):

        """
        Divider of K-fold cross image dataset.

        Parameters
        ----------
        K_group
            Number of fold cross.
        output_path
            Path of output folders.
        """

        rawimg = []
        rawimg.extend(glob.glob(os.path.join(self.record_dir, "*.png")))

        allcelltype = []
        for each in rawimg:
            allcelltype.append(each.split("/")[-1].split("-")[0])

        Dir = list(set(allcelltype))

        for i in range(K_group):
            os.makedirs(output_path + "/group" + str(i+1) + "/")
            os.makedirs(output_path + "/test" + str(i+1) + "/")
            for each in rawimg:
                shutil.copy(each, output_path + "/group" + str(i+1) + "/")
            imgsLib = []
            for n in Dir:
                os.makedirs(output_path + "/group" + str(i+1) + "/" + n + "/")
                os.makedirs(output_path + "/test" + str(i+1) + "/" + n + "/")
                imgsLib.extend(glob.glob(os.path.join(output_path + "/group" + str(i+1) + "/", n + "*.png")))
                for m in imgsLib:
                    shutil.move(m, output_path + "/group" + str(i+1) + "/" + n + "/")
                imgsLib = []
                imgsLib.extend(glob.glob(os.path.join(output_path + "/group" + str(i+1) + "/" + n + "/", "*.png")))
                np.random.shuffle(imgsLib)
                num = int(len(imgsLib)/K_group)
                for j in range(len(imgsLib[:num])):
                    shutil.move(imgsLib[j], output_path + "/test" + str(i+1) + "/" + n + "/")
                imgsLib = []





class Detection():

    def __init__(self):

        """
        Detection of spot images.

        """

        self.name = "Detection"



    def DetectWhiteRegion(self, input_path, grey_level, threshold):

        """
        Detection of spot images with lagre white regions.

        Parameters
        ----------
        input_path
            Path of input images.
        grey_level
            Grey-level threshold of the image.
        threshold
            Threshold of white regions.
        """
        
        imgsLib = []
        imgsLib.extend(glob.glob(os.path.join(input_path, "*.png")))
        
        if not os.path.exists(input_path + "/whiteimg/"):
            os.makedirs(input_path + "/whiteimg/")


        for each in imgsLib:
            img = Image.open(each)
            imgG = img.convert('L')

            rows = imgG.size[0]
            cols = imgG.size[1]

            q = 0
            for i in range(rows):
                for j in range(cols):
                    if imgG.getpixel((i, j)) > grey_level:
                        q += 1
                        
            if q/(rows*cols) > threshold:
                shutil.move(each, input_path + "/whiteimg/")


from setuptools import Command, find_packages, setup

__lib_name__ = "STASCAN"
__lib_version__ = "1.0.0"
__description__ = "An AI-driven method for enhanced cellular organizational map in spatial transcriptomics"
__url__ = "https://github.com/AbbyWY/STASCAN"
__author__ = "Ying Wu"
__author_email__ = "wuy@big.ac.cn"
__license__ = "MIT"
__requires__ = ["requests",]
__long_description__ = "Here we propose STASCAN, which both gene expression and the morphological information are simultaneously utilized to improve the cellular resolution of captured domains and even gap regions. Besides, STASCAN is further designed to enable cell-type predictions at subdivide-spot resolution and construction of the 3D spatial cell map from histology images alone."


setup(
    name = __lib_name__,
    version = __lib_version__,
    description = __description__,
    url = __url__,
    author = __author__,
    author_email = __author_email__,
    license = __license__,
    packages = ['STASCAN'],
    install_requires = __requires__,
    zip_safe = False,
    include_package_data = True,
    long_description = __long_description__
)



import os
import glob
import shutil
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axisartist.axislines import SubplotZero
import matplotlib.image as imgplt
#%matplotlib inline



import STASCAN.preparation as preparation
import STASCAN.model as model






class Module():

    def __init__(self):
        self.name = "Module"



    def UnseenSpot(self, result_dir, position, image, crop_size, prelabel_path,
                   strategies='joint', threshold_proportion=0.6, threshold_ration=1.5,
                   shuffle_scale=0.2, lr=1e-3, epochs=50, model_name="base_model.h5"):

        """
        Cell annotation for unseen spots.

        Parameters
        ----------
        result_dir
            Path of the output folder.
        position
            The position of raw spots. Standard output file for Space Ranger.
        image
            The original H&E staining images, used for Space Ranger previously.
        crop_size
            Pixel radius of spot images.
        prelabel_path
            Path of prelabel files.
        strategies: {'joint', 'single_proportion', 'single_ration'}
            Type of strategies to label the cell types for each spot.
        threshold_proportion
            The threshold of prior spot selection. Spots which the proportion of dominant cell types reached the threshold were selected as prior spots.
        threshold_ration
            The threshold of prior spot selection. Spots which the proportion of dominant cell types exceeded the given multiple of the proportion of secondary cell types were selected as prior spots.
        shuffle_scale
            Shuffle scale between training and testing sets.
        lr
            Learning rate for SGDOptimizer.
        epochs
            Number of total epochs in training.
        model_name
            Name of trained model
        """

        # Preparation
        step1 = preparation.Generator(result_dir, position, image, crop_size)
        prior_spot = step1.PriorSpot_generator(prelabel_path, strategies, threshold_proportion, threshold_ration)
        step1.RawSpot_generator()
        step1.ImputedSpot_generator()

        step2 = preparation.Shuffler(result_dir + "/PriorSpot/")
        step2.Dataset_divider(prior_spot, shuffle_scale)

        # Training
        step3 = model.BuildModel()
        traindata_path, testdata_path = result_dir + "/PriorSpot/train/", result_dir + "/PriorSpot/test/"
        n_class = len(set([x[-1] for x in prior_spot]))
        step3.base_model(traindata_path, testdata_path, result_dir, n_class, lr, epochs, model_name)

        # Prediction
        step4 = model.BuildPrediction(result_dir + "/Models/" + model_name, result_dir)
        step4.prediction(prior_spot, result_dir + "/ImputedSpot", result_dir + "/ImputedSpot/imputed_spot.txt", "Imputed")
        step4.prediction(prior_spot, result_dir + "/RawSpot", result_dir + "/ImputedSpot/adjust_raw_spot.txt", "Raw")



    def SubdividedSpot(self, result_dir, position, image, crop_size, prelabel_path,
                       strategies='single_proportion', threshold_proportion=0.2, threshold_ration=1.5,
                       shuffle_scale=0.2, lr=1e-3, epochs=50, model_name="base_model.h5"):

        """
        Cell annotation for subdivided spots.

        Parameters
        ----------
        result_dir
            Path of the output folder.
        position
            The position of raw spots. Standard output file for Space Ranger.
        image
            The original H&E staining images, used for Space Ranger previously.
        crop_size
            Pixel radius of spot images.
        prelabel_path
            Path of prelabel files.
        strategies: {'joint', 'single_proportion', 'single_ration'}
            Type of strategies to label the cell types for each spot.
        threshold_proportion
            The threshold of prior spot selection. Spots which the proportion of dominant cell types reached the threshold were selected as prior spots.
        threshold_ration
            The threshold of prior spot selection. Spots which the proportion of dominant cell types exceeded the given multiple of the proportion of secondary cell types were selected as prior spots.
        shuffle_scale
            Shuffle scale between training and testing sets.
        lr
            Learning rate for SGDOptimizer.
        epochs
            Number of total epochs in training.
        model_name
            Name of trained model
        """

        # Preparation
        step1 = preparation.Generator(result_dir, position, image, crop_size)
        prior_spot = step1.PriorSpot_generator(prelabel_path, strategies, threshold_proportion, threshold_ration)
        step1.RawSpot_generator()
        step1.SubSpot_generator()

        step2 = preparation.Shuffler(result_dir + "/SubSpot/prior_divided/")
        step2.Dataset_divider(prior_spot, shuffle_scale)

        # Training
        step3 = model.BuildModel()
        traindata_path, testdata_path = result_dir + "/SubSpot/prior_divided/train/", result_dir + "/SubSpot/prior_divided/test/"
        n_class = len(set([x[-1] for x in prior_spot]))
        step3.base_model(traindata_path, testdata_path, result_dir, n_class, lr, epochs, model_name)

        # Prediction
        step4 = model.BuildPrediction(result_dir + "/Models/" + model_name, result_dir)
        step4.prediction(prior_spot, result_dir + "/SubSpot/raw_divided/", result_dir + "/SubSpot/raw_divided/subloc.txt", "Subdivided")



    def UnseenSection(self, result_dir, position, image, crop_size, model_path, input_spot):

        """
        Cell annotation for unseen sections.

        Parameters
        ----------
        result_dir
            Path of the output folder.
        position
            The position of raw spots. Standard output file for Space Ranger.
        image
            The original H&E staining images, used for Space Ranger previously.
        crop_size
            Pixel radius of spot images.
        model_path
            Path of the trained model used for prediction.
        input_spot
            Spots in the dataset, which used to determine labels.
        """

        # Prepatation
        step1 = preparation.Generator(result_dir, position, image, crop_size)
        step1.RawSpot_generator()
        step1.ImputedSpot_generator()

        # Prediction
        step2 = model.BuildPrediction(model_path, result_dir)
        step2.prediction(input_spot, result_dir + "/ImputedSpot", result_dir + "/ImputedSpot/imputed_spot.txt", "Adjacent_imputed")
        step2.prediction(input_spot, result_dir + "/RawSpot", result_dir + "/ImputedSpot/adjust_raw_spot.txt", "Adjacent_raw")





class Optional():

    def __init__(self):

        self.name = "Optional"



    def SectionSpecificTraining(self, output_path, sections_path, imagetype, crop_size,
                                strategies='joint', threshold_proportion=0.2, threshold_ration=1.5, shuffle_scale=0.2,
                                lr1=1e-3, epoch1=50, lr2=1e-4, epoch2=50):

        """
        Section-specific training.

        Parameters
        ----------
        sections_path
            Path of the section folders.
        imagetype
            The type of original H&E staining images.
        crop_size
            Pixel radius of spot images.
        strategies: {'joint', 'single_proportion', 'single_ration'}
            Type of strategies to label the cell types for each spot.
        threshold_proportion
            The threshold of prior spot selection. Spots which the proportion of dominant cell types reached the threshold were selected as prior spots.
        threshold_ration
            The threshold of prior spot selection. Spots which the proportion of dominant cell types exceeded the given multiple of the proportion of secondary cell types were selected as prior spots.
        shuffle_scale
            Shuffle scale between training and testing sets.
        """
      
        allsection = []
        for each in os.listdir(sections_path):
            allsection.append(each)
            if not os.path.exists(output_path + each):
                os.makedirs(output_path + each)
                #os.makedirs(output_path + each + "/PriorSpot/")
        
        if not os.path.exists(output_path + "/PriorSpot/"):
            os.makedirs(output_path + "/PriorSpot/")
                
        allpriorspot = []
        for each in allsection:
            section_number = each.split('section')[-1]
            specific_output = output_path + each + "/"
            section_rawdata = sections_path + "/" + each + "/"
            Section_Generator = preparation.Generator(specific_output, section_rawdata + 'tissue_positions_list.csv', glob.glob(os.path.join(section_rawdata, "*."+imagetype))[0], crop_size)
            specific_prior_spot = Section_Generator.PriorSpot_generator(section_rawdata, strategies, threshold_proportion, threshold_ration)
            allpriorspot.extend(specific_prior_spot)
            imgsLib = []
            imgsLib.extend(glob.glob(os.path.join(specific_output + "/PriorSpot", "*.png")))
            for n in imgsLib:
                shutil.copyfile(n, output_path + "/PriorSpot/" + n.split("/")[-1].split(".")[0] + '_' + str(section_number) + ".png")
            specific_shuffler = preparation.Shuffler(specific_output + "/PriorSpot/")
            specific_shuffler.Dataset_divider(specific_prior_spot, shuffle_scale)

        section_shuffler = preparation.Shuffler(output_path + "/PriorSpot/")
        section_shuffler.Dataset_divider(allpriorspot, shuffle_scale)
        
        all_type = os.listdir(output_path + "/PriorSpot/train/")
        
        for each in allsection:
            for n in all_type:
                if not os.path.exists(output_path + each + "/" + "/PriorSpot/train/" + n):
                    shutil.copytree(output_path + "/PriorSpot/train/" + n + '/', output_path + each + "/" + "/PriorSpot/train/" + n)
                    shutil.copytree(output_path + "/PriorSpot/test/" + n + '/', output_path + each + "/" + "/PriorSpot/test/" + n)
                if not os.listdir(output_path + each + "/" + "/PriorSpot/test/" + n + "/"):
                    os.rmdir(output_path + each + "/" + "/PriorSpot/test/" + n)
                    shutil.copytree(output_path + each + "/" + "/PriorSpot/train/" + n, output_path + each + "/" + "/PriorSpot/test/" + n)
                    
        

        Base_model = model.BuildModel()
        n_class = len(set([x[-1] for x in allpriorspot]))
        Base_model.base_model(output_path + "/PriorSpot/train/", output_path + "/PriorSpot/test/", output_path, n_class, lr1, epoch1, model_name="base_model.h5")

        for each in allsection:
            specific_output = output_path + each + "/"
            specific_model = model.BuildModel()
            specific_model.finetuned_model(output_path + "/Models/base_model.h5", specific_output + "/PriorSpot/train/", specific_output + "/PriorSpot/test/", specific_output, lr2, epoch2, model_name="finetuned_model.h5")



    def PseudoLabelling(self, output_path, input_img, prediction_file, prediction_label, probability=0.90):

        """
        Pseudo labelling.

        Parameters
        ----------
        output_path
            Path of the output folder.
        input_img
            Path of predicted image folders.
        prediction_file
            The prediction files with detailed probabilities of cell types.
        prediction_label
            The label of predicted classes.
        probability
            The threshold of probability to select pseudo labels.
        """

        if not os.path.exists(output_path + "/PseudoSpot/"):
            os.makedirs(output_path + "/PseudoSpot/")

        prediction = []
        with open(prediction_file) as f:
            for line in f.readlines():
                temp = line.split()
                prediction.append(temp)

        pseudo_spot = []
        for each in prediction:
            value = [float(x) for x in each[1:-2]]
            celltype = prediction_label[value.index(max(value))]
            if max(value) >= probability:
                pseudo_spot.append([each[0], each[-2], each[-1], celltype, celltype])
        print("The number of pseudo spots : ", len(pseudo_spot))
        np.savetxt(output_path + "/PseudoSpot/pseudo_spot.txt", pseudo_spot, fmt='%s', delimiter='\t')

        for each in pseudo_spot:
            shutil.copy(os.path.join(input_img, each[0] + ".png"), os.path.join(output_path + "/PseudoSpot/", each[-1] + "-" + each[0] + ".png"))

        return pseudo_spot





class Visualization():

    def __init__(self):

        self.name = "Visualization"



    def EnhancedPlot(self, raw_spot, prior_spot, imputed_spot, image_path, label_color, output_path, pointsize=5, back_image=True):

        """
        Visualization of enhanced cell annotations.

        Parameters
        ----------
        raw_spot
            Path of the predicted raw spot file.
        prior_spot
            Path of the predicted prior spot file.
        imputed_spot
            Path of the predicted imputed spot file.
        image_path
            Path of the original H&E staining image.
        label_color
            Dictionary of label colors.
        output_path
            Path of the output files.
        pointsize
            Size of scatter points.
        back_image
            Shown of the original H&E staining image.
        """

        Image = imgplt.imread(image_path)
        h, w, _ = Image.shape
        colors = label_color

        rawspot = {}
        with open(raw_spot) as f:
            for line in f.readlines():
                temp = line.split()
                rawspot[temp[0]] = temp[2:]

        priorspot = {}
        with open(prior_spot) as f:
            for line in f.readlines():
                temp = line.split()
                priorspot[temp[0]] = [temp[-1], temp[1], temp[2]]

        imputedspot = {}
        with open(imputed_spot) as f:
            for line in f.readlines():
                temp = line.split()
                imputedspot[temp[0]] = temp[2:]

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)

        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)
        ax.set_xticks([])
        ax.set_yticks([])

        if back_image == True:
            plt.imshow(Image)

        for each in rawspot.keys():
            plt.scatter(float(rawspot[each][1]), float(rawspot[each][2]), c='white', marker='o', s=pointsize+2, edgecolors='black', linewidths=0.1, alpha=0.3)
        for each in priorspot.keys():
            plt.scatter(float(rawspot[each][1]), float(rawspot[each][2]), c=colors[priorspot[each][0]], marker='o', s=pointsize)
        plt.savefig(output_path + "/PriorPlot.pdf")

        for each in rawspot.keys():
            if each not in priorspot.keys():
                plt.scatter(float(rawspot[each][1]), float(rawspot[each][2]), c=colors[rawspot[each][0]], marker='o', s=pointsize)
        for each in imputedspot.keys():
            plt.scatter(float(imputedspot[each][1]), float(imputedspot[each][2]), c=colors[imputedspot[each][0]], marker='o', s=pointsize)
        plt.savefig(output_path + "/EnhancedPlot.pdf")



    def SubResolutionPlot(self, raw_spot, sub_spot, image_path, label_color, output_path, pointsize=5, back_image=True):

        """
        Visualization of subdivided cell annotations.

        Parameters
        ----------
        raw_spot
            Path of the predicted raw spot file.
        sub_spot
            Path of the predicted subdivided spot file.
        image_path
            Path of the original H&E staining image.
        label_color
            Dictionary of label colors.
        output_path
            Path of the output files.
        pointsize
            Size of scatter points.
        back_image
            Shown of the original H&E staining image.
        """

        Image = imgplt.imread(image_path)
        h, w, _ = Image.shape
        colors = label_color

        rawspot = {}
        with open(raw_spot) as f:
            for line in f.readlines():
                temp = line.split()
                rawspot[temp[0]] = temp[2:]

        subspot = {}
        with open(sub_spot) as f:
            for line in f.readlines():
                temp = line.split()
                subspot[temp[0]] = temp[2:]
                
        #%matplotlib inline

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)

        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)
        ax.set_xticks([])
        ax.set_yticks([])

        if back_image == True:
            plt.imshow(Image)

        for each in rawspot.keys():
            plt.scatter(float(rawspot[each][1]), float(rawspot[each][2]), c=colors[rawspot[each][0]], marker='o', s=pointsize)
        plt.savefig(output_path + "/RawResolution.pdf")
        #plt.show()
        #plt.clf()

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)

        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)
        ax.set_xticks([])
        ax.set_yticks([])

        if back_image == True:
            plt.imshow(Image)

        for each in subspot.keys():
            plt.scatter(float(subspot[each][1]), float(subspot[each][2]), c=colors[subspot[each][0]], marker='o', s=int(pointsize/2))
        plt.savefig(output_path + "/SubResolutionPlot.pdf")
        #plt.clf()



    def AdjacentSectionPlot(self, raw_spot, imputed_spot, image_path, label_color, output_path, pointsize=5, back_image=True):

        """
        Visualization of adjacent section cell annotations.

        Parameters
        ----------
        raw_spot
            Path of the predicted raw spot file.
        imputed_spot
            Path of the predicted imputed spot file.
        image_path
            Path of the original H&E staining image.
        label_color
            Colors of class labels.
        output_path
            Path of the output files.
        pointsize
            Size of scatter points.
        back_image
            Shown of the original H&E staining image.
        """

        Image = imgplt.imread(image_path)
        h, w, _ = Image.shape
        colors = label_color

        rawspot = {}
        with open(raw_spot) as f:
            for line in f.readlines():
                temp = line.split()
                rawspot[temp[0]] = temp[2:]

        imputedspot = {}
        with open(imputed_spot) as f:
            for line in f.readlines():
                temp = line.split()
                imputedspot[temp[0]] = temp[2:]

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)

        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)
        ax.set_xticks([])
        ax.set_yticks([])

        if back_image == True:
            plt.imshow(Image)

        for each in rawspot.keys():
            plt.scatter(float(rawspot[each][1]), float(rawspot[each][2]), c=colors[rawspot[each][0]], marker='o', s=pointsize)
        for each in imputedspot.keys():
            plt.scatter(float(imputedspot[each][1]), float(imputedspot[each][2]), c=colors[imputedspot[each][0]], marker='o', s=pointsize)
        plt.savefig(output_path + "/AdjacentSectionPlot.pdf")









import matplotlib.pyplot as plt
from mpl_toolkits.axisartist.axislines import SubplotZero
from sklearn.metrics import confusion_matrix
from skimage.metrics import structural_similarity as ssim
from itertools import cycle
from sklearn.metrics import roc_curve, auc
from sklearn.preprocessing import label_binarize
import cv2
import numpy as np
import itertools
import json
import os
import glob
import matplotlib.image as imgplt





class Metric():

    def __init__(self):

        """
        Metrics.

        """

        self.name = "Metric"



    def Loss_Accuracy_Curve(self, input_file, output_path):

        """
        Visualization of loss curve and accuracy curve.

        Parameters
        ----------
        input_file
            Path of input files.
        output_path
            Path of output files.
        """

        with open(input_file) as f:
            loss = list(map(float, f.readlines()[0].split()))
        with open(input_file) as f:
            val_loss = list(map(float, f.readlines()[1].split()))
        with open(input_file) as f:
            accuracy = list(map(float, f.readlines()[2].split()))
        with open(input_file) as f:
            val_accuracy = list(map(float, f.readlines()[3].split()))

        x = range(1, len(loss)+1)

        plt.figure(figsize=(12, 5))

        # Loss Curve
        plt.subplot(1, 2, 1)
        plt.xlim(0, len(loss)+1)
        plt.ylim(0, max(max(loss), max(val_loss))+1)
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.xticks(np.arange(0, len(loss)+1, 10))

        plt.plot(x, loss, c='#228B22', linestyle=':', linewidth=1, label='Train')
        plt.scatter(x, loss, c='#228B22', s=1, marker='o')
        plt.text(x[-1]+1, loss[-1], '%.2f' % loss[-1], ha='center', va='bottom', fontsize=5)

        plt.plot(x, val_loss, c='#1874CD', linestyle=':', linewidth=1, label='Test')
        plt.scatter(x, val_loss, c='#1874CD', s=1, marker='o')
        plt.text(x[-1], val_loss[-1], '%.2f' % val_loss[-1], ha='center', va='bottom', fontsize=5)

        plt.legend(loc='upper right', fontsize=5)

        # Accuracy Curve
        plt.subplot(1, 2, 2)
        plt.xlim(0, len(accuracy) + 1)
        plt.ylim(0, 1)
        plt.xlabel('Epoch')
        plt.ylabel('Accuracy')
        plt.xticks(np.arange(0, len(accuracy) + 1, 10))

        plt.plot(x, accuracy, c='#EE7600', linestyle=':', linewidth=1, label='Train')
        plt.scatter(x, accuracy, c='#EE7600', s=1, marker='o')
        plt.text(x[-1] + 1, accuracy[-1] + 0.02, '%.2f' % (accuracy[-1] * 100) + "%", ha='center', va='bottom', fontsize=5)

        plt.plot(x, val_accuracy, c='#CD2626', linestyle=':', linewidth=1, label='Test')
        plt.scatter(x, val_accuracy, c='#CD2626', s=1, marker='o')
        plt.text(x[-1], val_accuracy[-1] - 0.06, '%.2f' % (val_accuracy[-1] * 100) + "%", ha='center', va='bottom', fontsize=5)

        plt.legend(loc='lower left', fontsize=5)

        plt.savefig(output_path + "/Accr-Loss.pdf")



    def plot_confusion_matrix(self, cm, classes, output_fig):

        """
        Plotting confusion matrix.

        Parameters
        ----------
        cm
            Inputted confusion matrix.
        classes
            List of predicted classes.
        output_fig
            Path of output files.
        """
        
        plt.figure(figsize=(5, 5))

        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Reds)
        plt.title('The Recall Value of Classes')
        plt.colorbar()
        tick_marks = np.arange(len(classes))
        plt.xticks(tick_marks, classes, rotation=45)
        plt.yticks(tick_marks, classes)
        thresh = cm.max() / 2.
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            plt.text(j, i, '{:.2f}'.format(cm[i, j]), horizontalalignment="center", color="white" if cm[i, j] > thresh else "black")
            #plt.text(j, i + 0.3, '(' + '{:,}'.format(cm[i, j]) + ')', horizontalalignment="center", color="black", fontsize=6)
        plt.tight_layout()
        plt.ylabel('Truth')
        plt.xlabel('Prediction')
        plt.savefig(output_fig)



    def Confusion_Matrix(self, image_path, predicted_path, label, output_fig):

        """
        Visualization of confusion matrix.

        Parameters
        ----------
        image_path
            Path of image folder.
        predicted_path
            Path of predicted files.
        label
            List of predicted labels.
        output_fig
            Path of output files.
        """

        imgsLib = []
        for n in label:
            imgsLib.extend(glob.glob(os.path.join(image_path + "/" + n + "/", "*.png")))

        Labels = {}
        for each in imgsLib:
            Labels[each.split("/")[-1].split(".")[0].split("-")[1]+'-'+each.split("/")[-1].split(".")[0].split("-")[2]] = [each.split("/")[-1].split("-")[0]]

        with open(predicted_path) as f:
            for line in f.readlines():
                temp = line.split()
                if temp[0] in Labels.keys():
                    Labels[temp[0]].append(temp[2])

        for each in Labels:
            if len(Labels[each]) != 2:
                print("Error!")
                print(str(each))


        label_true, label_pred = [], []
        for each in Labels.keys():
            label_true.append(Labels[each][0])
            label_pred.append(Labels[each][1])

        cm = confusion_matrix(label_true, label_pred, labels=label)
        self.plot_confusion_matrix(cm, label, output_fig+"/Confusion_Matrix.pdf")



    def ROC_Curve(self, image_path, predicted_path, labels, color, output_fig):

        """
        Visualization of ROC curve.

        Parameters
        ----------
        image_path
            Path of image folder.
        predicted_path
            Path of predicted files with detailed probabilities.
        labels
            List of predicted labels.
        color
            List of colors for labels.
        output_fig
            Path of output files.
        """

        digtal_label, label_digtal = {}, {}
        labels.sort()
        for i in range(len(labels)):
            digtal_label[i] = labels[i]
            label_digtal[labels[i]] = i

        imgsLib = []
        for n in labels:
            imgsLib.extend(glob.glob(os.path.join(image_path + "/" + n + "/", "*.png")))

        Label_true = {}
        for each in imgsLib:
            Label_true[each.split("/")[-1].split(".")[0].split("-")[1] + '-' + each.split("/")[-1].split(".")[0].split("-")[2]] = [each.split("/")[-1].split("-")[0]]

        label_true_digtal = []
        for each in Label_true:
            label_true_digtal.append(label_digtal[Label_true[each][0]])

        y_test = label_binarize(np.array(label_true_digtal), classes=list(range(0, len(labels))))
        n_classes = y_test.shape[1]

        with open(predicted_path) as f:
            for line in f.readlines():
                temp = line.split()
                if temp[0] in Label_true.keys():
                    temp[1:-2] = [float(k) for k in temp[1:-2]]
                    Label_true[temp[0]].extend(temp[1:-2])

        p_type_detail = []
        for each in Label_true:
            p_type_detail.append(Label_true[each][1:])

        y_score = np.array(p_type_detail)

        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(n_classes):
            fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
            roc_auc[i] = auc(fpr[i], tpr[i])

        lw = 2
        plt.figure(figsize=(5, 5))

        colors = cycle(color)
        for i, color, label in zip(range(n_classes), colors, cycle(list(label_digtal.keys()))):
            plt.plot(fpr[i], tpr[i], color=color, lw=2, label=label + '(AUC = {1:0.3f})' ''.format(i, roc_auc[i]))

        plt.plot([0, 1], [0, 1], 'k--', lw=lw)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.legend(loc="lower right", fontsize=7)
        plt.savefig(output_fig)



    def SSIM(self, imfil1, imfil2):

        """
        Calculation of SSIM.

        Parameters
        ----------
        imfil1
            Path of image 1.
        imfil2
            Path of image 2.
        """

        img1 = cv2.imread(imfil1)
        (h, w) = img1.shape[:2]
        img2 = cv2.imread(imfil2)
        resized = cv2.resize(img2, (w, h))
        (h1, w1) = resized.shape[:2]
        img1 = cv2.cvtColor(img1, cv2.COLOR_BGR2RGB)
        img2 = cv2.cvtColor(resized, cv2.COLOR_BGR2RGB)

        return ssim(img1, img2, multichannel=True)





class Check():

    def __init__(self):

        """
        Visualization of pre-labelling or imputing.

        """

        self.name = "Check"



    def Check_PriorSpot(self, adjust_raw_spot, prior_spot, image_path, label_color, output_path, pointsize=5, back_image=False):

        """
        Visualization of pre-labelling.

        Parameters
        ----------
        adjust_raw_spot
            Path of the adjusted raw spot file.
        prior_spot
            Path of the prior spot file.
        image_path
            Path of the original H&E staining image.
        label_color
            Dictionary of label colors.
        output_path
            Path of the output files.
        pointsize
            Size of scatter points.
        back_image
            Shown of the original H&E staining image.
        """

        Image = imgplt.imread(image_path)
        h, w, _ = Image.shape
        colors = label_color

        rawspot = {}
        with open(adjust_raw_spot) as f:
            for line in f.readlines():
                temp = line.split()
                rawspot[temp[-1]] = temp[0:2]

        priorspot = {}
        select = []
        with open(prior_spot) as f:
            for line in f.readlines():
                temp = line.split()
                priorspot[temp[0]] = [temp[-1], temp[1], temp[2]]
                select.append(temp)

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)

        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)
        ax.set_xticks([])
        ax.set_yticks([])

        if back_image == True:
            plt.imshow(Image)

        for each in rawspot.keys():
            plt.scatter(float(rawspot[each][0]), float(rawspot[each][1]), c='white', marker='o', s=pointsize+2, edgecolors='black', linewidths=0.1)
        for each in priorspot.keys():
            plt.scatter(float(priorspot[each][1]), float(priorspot[each][2]), c=colors[priorspot[each][0]], marker='o', s=pointsize)
        plt.savefig(output_path + "/Check_PriorSpot.pdf")

        selectnum = list(set([x[-1] for x in select]))
        Label_selectnum = {}
        for each in selectnum:
            Label_selectnum[each] = 0
        for each in select:
            Label_selectnum[each[-1]] = Label_selectnum[each[-1]] + 1
        info_select = json.dumps(Label_selectnum, sort_keys=False)

        f = open(output_path + "stat.txt", "w")
        f.write(info_select)
        f.close()



    def Check_ImputedSpot(self, adjust_raw_spot, imputed_spot, image_path, output_path, pointsize=5, back_image=False):

        """
        Visualization of imputing.

        Parameters
        ----------
        adjust_raw_spot
            Path of the adjusted raw spot file.
        imputed_spot
            Path of the imputed spot file.
        image_path
            Path of the original H&E staining image.
        output_path
            Path of the output files.
        pointsize
            Size of scatter points.
        back_image
            Shown of the original H&E staining image.
        """

        Image = imgplt.imread(image_path)
        h, w, _ = Image.shape

        rawspot = {}
        with open(adjust_raw_spot) as f:
            for line in f.readlines():
                temp = line.split()
                rawspot[temp[-1]] = temp[0:2]

        imputedspot = {}
        with open(imputed_spot) as f:
            for line in f.readlines():
                temp = line.split()
                imputedspot[temp[-1]] = temp[0:2]

        fig = plt.figure(figsize=(10, 10))
        ax = SubplotZero(fig, 1, 1, 1)
        fig.add_subplot(ax)

        ax.set_xlim(left=0, right=int(max(h, w)))
        ax.set_ylim(bottom=int(max(h, w)), top=0)
        ax.set_xticks([])
        ax.set_yticks([])

        if back_image == True:
            plt.imshow(Image)

        for each in rawspot.keys():
            plt.scatter(float(rawspot[each][0]), float(rawspot[each][1]), c='red', marker='o', s=pointsize)
        for each in imputedspot.keys():
            plt.scatter(float(imputedspot[each][0]), float(imputedspot[each][1]), c='blue', marker='o', s=pointsize)
        plt.savefig(output_path + "/Check_ImputedSpot.pdf")

